#!/usr/bin/python
# autoqa-autotest: script to handle incoming requests from watchers and kick
#                  off appropriate autotest jobs.
#
# Copyright 2009, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Author: Will Woods <wwoods@redhat.com>

import os
import sys
import optparse
import tempfile
import copy
import fnmatch
import traceback
import subprocess
import logging
import autoqa.logger
from autoqa.config import Config

# don't use proxy when scheduling jobs
autoqa.config.use_proxy(False)

def prep_controlfile(controlfile, extradata, arch):
    '''Returns the name of a temporary file containing the given controlfile,
    prepended with key='value' lines for each item in extradata and an
    'autoqa_conf' variable which holds the contents of the autoqa configfile.
    '''
    # the jobtag must be evaluated on every single testrun (since it changes :-))
    # so line 'jobtag = %(jobtag)s' is added to the [job] part of the autoqa_conf
    # string, and then is replaced with job.tag when the code is executed

    controldata = open(controlfile)
    (fd, name) = tempfile.mkstemp(prefix='autoqa-control.')
    os.write(fd, '# -*- coding: utf-8 -*-\n\n')

    extradata['arch'] = arch or 'noarch'
    os.write(fd, "autoqa_args = %r" % extradata)
    os.write(fd,'\n')
    os.write(fd, "autoqa_args['config'] = {'jobtag' : job.tag}")
    os.write(fd,'\n\n')
    os.write(fd,controldata.read())
    os.close(fd)
    return name

def maybe_call(cmdlist, dryrun=False, logerror=False):
    '''Call a command
       @param cmdlist list of call elements
       @param dryrun if True, just print out the command and return
       @param logerror if True, log command output if it fails (this pipes the
              output to a variable, so it's not printed out in real-time)
       @return return code of the executed command
    '''
    if dryrun:
        print ' '.join(cmdlist)
        return 0

    stdout = subprocess.PIPE if logerror else None

    try:
        process = subprocess.Popen(cmdlist, stdout=stdout,
                                   stderr=subprocess.STDOUT)
        process.wait()

        if logerror:
            output = process.stdout.read().rstrip()
            if process.returncode == 0:
                print output
            else:
                logging.error("Command failed: %s" % ' '.join(cmdlist))
                logging.error(output)

        return process.returncode

    except OSError, e:
        logging.error("Command failed: %s\n%s" % (' '.join(cmdlist), e))
        return 1

def schedule_job(controlfile, required_arch='noarch', email=None, name=None,
                 dryrun=False, labels=[]):
    cmd = ['/usr/bin/atest', 'job', 'create']
    if email:
        cmd += ['-e', email]

    # don't reboot the client before or after the test runs
    cmd += ["--reboot_before=never", "--reboot_after=never"]

    # run on required system architecture
    required_arch = required_arch or 'noarch'
    # for 'i[3-6]86' arch we have 'i386' autotest label, convert it
    if fnmatch.fnmatch(required_arch, 'i?86'):
        required_arch = 'i386'
    cmd += ['-m', '*%s' % required_arch]

    # schedule against additional labels, like distro label ('fc16')
    if labels:
        cmd += ['-d', ','.join(labels)]

    cmd += ['-f', controlfile]
    cmd.append(name) # job name

    return maybe_call(cmd, dryrun, logerror=True)

def run_test_locally(controlfile, name=None, dryrun=False):
    cmd = ['/usr/share/autotest/client/bin/autotest', '--verbose']
    if name:
        cmd += ['-t', name]
    cmd.append(controlfile)
    return maybe_call(cmd, dryrun)

def eval_test_vars(test, test_vars):
    '''Take a test's control.autoqa file, have the test_vars argument as the
    input data, evaluate the file, and return the modified test_vars.
    Arguments:
        * test - name of the test for which control.autoqa should be executed
        * test_vars - dictionary with test variables that will be used for input
    Returns: dictionary with test variables that have been evaluated (and
    probably modified) by test's control.autoqa file
    '''
    cfile = open(os.path.join(Config.paths.testdir, test, 'control.autoqa'))
    # we need to deepcopy test_vars, so we don't change the input argument at all
    # (there are nested structures)
    vars = copy.deepcopy(test_vars)
    # change dir to the dir of the current test, it may need it
    olddir = os.getcwdu()
    os.chdir(os.path.dirname(os.path.abspath(cfile.name)))
    # execute the file
    exec cfile in vars
    # change dir back
    try:
        os.chdir(olddir)
    except:
        # this might not work (e.g. when running command using sudo), but
        # it's not that important, so just be silent about it
        pass
    cfile.close()
    # leave only those keys there were defined before, delete all other keys
    # (there could have appeared some new ones we don't need, like builtins
    # or temporary variables)
    for key in vars.keys():
        if not key in test_vars.keys():
            del vars[key]
    return vars

def prep_test_vars(event, archs, autoqa_args):
    '''Prepare variables that should be redirected to a test's control.autoqa
    file, so the test can then decide whether to run and how.
    Arguments:
        event - name of the event that calls this test
        archs - list of architectures to be executed on
        autoqa_args - dictionary of autoqa arguments that will be given to
        test's control file
    Returns a dictionary of all the test variables, including autoqa_args
    as one of its items.
    '''
    test_vars = {}
    test_vars['autoqa_args'] = autoqa_args
    test_vars['event'] = event
    test_vars['archs'] = archs
    test_vars['labels'] = []
    test_vars['execute'] = False
    return test_vars

def resolve_tests(opts, autoqa_args, event, parser):
    # Evaluate control.autoqa file for every test to get a list of tests to execute
    tests = [test for test in os.listdir(Config.paths.testdir)]
    default_test_vars = prep_test_vars(event, opts.arch, autoqa_args)
    test_vars = {} # dict of test->its test vars
    for test in tests[:]:
        try:
            test_vars[test] = eval_test_vars(test, default_test_vars)
        except StandardError, e:
            logging.error("Can't evaluate test '%s': %s" % (test, e))
            traceback.print_tb(sys.exc_traceback)
            tests.remove(test)
    testlist = [test for test,vars in test_vars.iteritems()
                if vars['execute'] == True]

    # Allow testlist user override
    # User may force some test to run even though it wouldn't be run by default.
    # This is useful for example for helloworld test.
    if opts.test:
        for t in opts.test:
            if t not in tests:
                parser.error('Unknown test: %s' % t)
        testlist = opts.test
    return (testlist, test_vars)

def parse_params():
    # Sanity check our installation
    if not os.path.isdir(Config.paths.eventdir):
        logging.error("Can't find events in %s. Check your installation." %
                      Config.paths.eventdir)
        sys.exit(1)
    # known events = dirs in eventdir
    # FIXME - would be nice to have some way to determine what makes a event valid
    known_events = [d for d in os.listdir(Config.paths.eventdir)
                    if os.path.isdir(os.path.join(Config.paths.eventdir, d))]

    # Set up the option parser
    parser = optparse.OptionParser(usage="%prog EVENT [options] ...",
        add_help_option=False)

    parser.add_option('-h', '--help', action='help',
        help='show this help message (or event help message if EVENT given) and \
    exit')

    parser.add_option('-a', '--arch', action='append', default=[],
        help="arch to run the test(s) on; can be used multiple times; by default \
    'noarch' arch is used")

    parser.add_option('-t', '--test', action='append',
        help="run only the given test(s) instead of all relevant ones; can be used \
    multiple times; if you specify a test that wouldn't be run by default it will \
    be forced to run")

    # XXX --skiptest/--exclude?
    parser.add_option('--keep-control-file', action='store_true',
        help='do not delete generated control files')

    parser.add_option('--dryrun', '--dry-run', action='store_true', dest='dryrun',
        help='do not actually execute commands, just show what would be done \
    (implies --keep-control-file)')

    parser.add_option('--local', action='store_true', dest='local',
        help='do not schedule jobs - run test(s) directly on the local machine')

    parser.add_option('-l', '--list-tests', action='store_true', dest='listtests',
        help='list the tests for the given event - do not run any tests')

    parser.add_option('--autotest-server', action='store', default=None,
        help='sets the autotest-server hostname used for creating URLs to results;\
    hostname of the local machine is used by default')

    autoqa.logger.add_options(parser)

    # Read and validate the event name
    # Check for no args, or just -h/--help
    if len(sys.argv) == 1 or sys.argv[1] in ('-h', '--help'):
        parser.print_help()
        logging.error('No event given.\nKnown events: %s' % ' '.join(known_events))
        parser.exit()
    event_name = sys.argv[1]
    if not event_name in known_events:
        parser.error('Bad event name "%s".\nKnown events: %s' % (event_name,
                     ' '.join(known_events)))
    eventdir = os.path.join(Config.paths.eventdir, event_name)
    # Hooray, we have a good eventdir. Pull in the event-specific parsing code.
    sys.path.append(eventdir)
    import event
    event.extend_parser(parser)

    # And now we really parse the commandline.
    (opts, args) = parser.parse_args()
    args.pop(0) # dump event name

    # Set up a logger
    autoqa.logger.setup_logger(file_logger=not opts.dryrun, opts=opts)

    # Run the tests locally, or schedule them through autotest?
    run_local = (opts.local or Config.general.local)

    if not opts.arch or run_local:
        opts.arch = ['noarch']
    # it doesn't make sense to have 'noarch' and some other arch specified, but
    # some watcher may still provide us with such combination
    # if this is the case, just delete the 'noarch' item
    while 'noarch' in opts.arch and len(opts.arch) > 1:
        opts.arch.remove('noarch')

    # Override autotest_server if required
    if opts.autotest_server:
        Config.general.autotest_server = opts.autotest_server

    # Ask event to determine all arguments needed for tests. These variables will
    # be then written into the control file as autoqa_args dictionary. The event can
    # also call parser.error if it find out that not all options are correctly
    # populated.
    autoqa_args = event.process_testdata(parser, opts, args)
    # The test may also need the very name of the event that executes it
    autoqa_args['event'] = event_name
    return (opts, args, autoqa_args, event_name, run_local, parser)

def main():
    (opts, args, autoqa_args, event, run_local, parser) = parse_params()

    (testlist, test_vars) = resolve_tests(opts, autoqa_args, event, parser)

    # Print testlist, if requested
    if opts.listtests:
        logging.info(' '.join(testlist))
        sys.exit(0)

    # We're ready to run/queue tests now.
    exit_status = 0
    for test in testlist:
        template = os.path.join(Config.paths.testdir, test, 'control')

        for arch in test_vars[test]['archs']:
            try:
                control = prep_controlfile(template,
                                           test_vars[test]['autoqa_args'], arch)
            except IOError, e:
                logging.warn("Could not process control file for %s: %s" %
                             (test, str(e)))
                continue
            testname='%s:%s.%s' % (event, test, arch)
            email = Config.notifications.notification_email

            if run_local:
                retval = run_test_locally(control, name=testname,
                                          dryrun=opts.dryrun)
            else:
                retval = schedule_job(control, email=email, name=testname,
                                      required_arch=arch, dryrun=opts.dryrun,
                                      labels=test_vars[test]['labels'])
                if retval != 0:
                    logging.error("Failed to schedule job %s" % testname)
                    exit_status = 10

        if opts.keep_control_file or opts.dryrun:
            logging.info("Keeping %s at user request" % control)
        else:
            os.remove(control)

    sys.exit(exit_status)

if __name__ == "__main__":
    main()
